//-----------------------------------------------------------
//
// SSE2030: Introduction to Computer Systems (Fall 2012)
//
// Skeleton code for PA#1
// 
// September 3, 2012.
// CSLab, Sungkyunkwan University
//
// Submit Date  : 2012-09-05
// Student Id   : 2012711289
// Student Name : Jinhyuk Lee
//
// Arguments : char* fileName - Novel text file path
//             int numExceptWords - Number of except words
//             exceptWords - Except word list array
// 
// Return    : Character pointer array of top-10 words
//-----------------------------------------------------------

/** 
 *  @file       topk.c
 *  @ingroup    
 *  @brief      This file implements counting word and ranking
 *  @author     JH.LEE 
 *  @date       2012/09/04
 *  @remark
 */


//***************************************************************
//  Header file inclusion
#include <stdio.h>
#include "malloc.h"
#include "topk.h"
#include "list.h"

//***************************************************************
// Global variable definition

//***************************************************************
// Imported variable definition

//***************************************************************
//  Imported function prototype

//***************************************************************
//  Local macro definition
//#define __EANABLE_ASSERT

#if defined(__EANABLE_ASSERT)
#define TOPK_DBG_ASSERT(condition)  do { if(!(condition)){ while (1);} } while (0);
#else
#define TOPK_DBG_ASSERT(condition) 
#endif

#define INIT_WORD_POOL(pstWord)     do {   (pstWord)->nTotalRefCnt = 0; \
                                            INIT_LIST_HEAD((&((pstWord)->stSlotList))); \
                                        } while (0);

//***************************************************************
// Local constant definition
#define MAX_NUM_STR_BUF_SIZE            (4096)
#define MAX_NUM_WORD_BUF_SIZE           (48) // pnuemonoultramicroscopicsilicovolcanoconiosis is the longest word
#define MIN_NUM_WORD_LENGTH             (3)
#define NUM_STRING_CHK_UNIT             (MAX_NUM_WORD_BUF_SIZE >> 2)

#define NUM_HASH_BUCKET                 (256) // prime number
#define MASK_BIT_HASHKEY                (0xFF)  
#define SHFT_BIT_HASHKEY                (8)  

//***************************************************************
// Local type definition
typedef unsigned char                   UINT8;  
typedef char                            INT8;   
typedef unsigned short                  UINT16; 
typedef short                           INT16;  
typedef unsigned long int               UINT32; 
typedef long int                        INT32;  
typedef unsigned long long int          UINT64; 
typedef long long int                   INT64;  
typedef UINT32                          BOOL32;

#define TRUE32                          (0x00000001)
#define FALSE32                         (0x00000000)
#define PUBLIC                          
#define PRIVATE                         static
#define VOID                            void


/**
 *  @brief WORD_LIST
 */ 
typedef struct {
    UINT8               astrWordBuf[MAX_NUM_WORD_BUF_SIZE]; //48 
    UINT16              nTotalRefCnt;                       //2
    struct list_head    stSlotList;                         //8
} WORD_ENTRY; // 64byte

/**
 *  @WORD_POOL Pool
 */ 
typedef struct {
    struct list_head    *pstSlotHead;
    UINT16              nEntryCnt;
} HASH_ENTRY;

/**
 *  @WORD_POOL Pool
 */ 
typedef struct {
    HASH_ENTRY astTopKBucket[NUM_HASH_BUCKET];
} WORD_POOL;



//***************************************************************
// Local variable definition
struct list_head    *gpstRankListHead = NULL;
UINT32              gnTotalWordCnt = 0;
UINT32              gnTotalWordPoolCnt = 0;
WORD_POOL           gstWordPool;

//***************************************************************
// Local(static) function declaration
PRIVATE BOOL32      _CheckCriteria(UINT8 *str);
PRIVATE VOID        _PushWordPool(WORD_ENTRY *pstWordEntity);
PRIVATE BOOL32      _CompareString(UINT32 *pTargetStr, UINT32 *pRefStr);
PRIVATE VOID        _CopyString(UINT8 *pDstStr, UINT8 *pSrcStr);
PRIVATE UINT16       _GenHashKey(WORD_ENTRY *pstWordEntity);
PRIVATE VOID        _InsertToHashBucket(WORD_ENTRY *pstWordEntity, UINT16 nBucketIdx);
PRIVATE WORD_ENTRY*  _GetWordEntity(VOID);
PRIVATE VOID        _ReleaseWordPool(VOID);
PRIVATE VOID        _UpdateTopKList(UINT32 nNumK, struct wordcount *topk);
//***************************************************************
// Code implementation

/**
 *  @fn      PUBLIC INT32 getTopk(char* fileName, int numk, struct wordcount *topk) 
 *  @param   fileName
 *  @param   numk
 *  @param   topk
 *  @return  INT32 total word counts
 *  @par     This function implements reporting rank of work reference frequency
 *  @remark 
 */ 
PUBLIC 
INT32 getTopK(INT8* fileName, INT32 numk, struct wordcount *topk) 
{
    UINT8       anStrBuf[MAX_NUM_STR_BUF_SIZE];
    UINT32      nReadStreamCnt = 0;
    UINT32      nIter = 0;
    UINT32      nCurStrBufIdx = 0;
    UINT32      nCurWordBufIdx = 0;
    WORD_ENTRY   *pstWordEntity = NULL;

    FILE    *ptext;

    // Open text file
    ptext = fopen(fileName, "rt");

    if (NULL == ptext)
    {
        printf("file open error! \n");
    }
    
    // Init stream buffer
    for (nIter = 0; nIter < MAX_NUM_STR_BUF_SIZE; nIter++)
    {
        anStrBuf[nIter] = 0x0; 
    }

    // Init WORD_POOL Table
    for (nIter = 0; nIter < NUM_HASH_BUCKET; nIter++)
    {
        gstWordPool.astTopKBucket[nIter].pstSlotHead = NULL;
        gstWordPool.astTopKBucket[nIter].nEntryCnt = 0;
    }

    while(!feof(ptext))
    {
        // Get Stream
        nReadStreamCnt = fread(anStrBuf, sizeof(UINT8), MAX_NUM_STR_BUF_SIZE, ptext);
       
        // Init string buffer index
        nCurStrBufIdx = 0;
        
        // Parsing word
        do
        {
            if (NULL == pstWordEntity)
            {
                pstWordEntity = _GetWordEntity();
            }
            
            // Check word parsing 
            if ((anStrBuf[nCurStrBufIdx] == ' ') ||
                (anStrBuf[nCurStrBufIdx] == '\n')||
                (anStrBuf[nCurStrBufIdx] == '\t')) 
            {
                // Insert NULL
                pstWordEntity->astrWordBuf[nCurWordBufIdx] = '\0';
                //anWordBuf[nCurWordBufIdx] = '\0';
                    
                // Check word length, warn adding NULL character to count total word count
                if (nCurWordBufIdx >= MIN_NUM_WORD_LENGTH)
                {
                    // Sorting Ranking and update word count
                    _PushWordPool(pstWordEntity);
                    gnTotalWordCnt++;

                    // Clear current word entity pointer because the entity was inserted word pool
                    pstWordEntity = NULL;
                }

                // Init word buffer index
                nCurWordBufIdx = 0;
            }
            else
            {
                if (TRUE32 == _CheckCriteria(&anStrBuf[nCurStrBufIdx]))
                {
                    // Insert character to word buffer
                    pstWordEntity->astrWordBuf[nCurWordBufIdx] = anStrBuf[nCurStrBufIdx];
                    nCurWordBufIdx++;
                    TOPK_DBG_ASSERT(nCurWordBufIdx < MAX_NUM_WORD_BUF_SIZE);
                }
                else
                {
                    /* NULL */
                }
            }
        } while (++nCurStrBufIdx < nReadStreamCnt);
    } 

    // Release non parsed garbage word entity
    if (NULL != pstWordEntity)
    {
        free(pstWordEntity);
    }

    // update topk
    _UpdateTopKList((UINT32)numk, topk);
    
    // free except remained word pool
    _ReleaseWordPool();

    // Close text file
    fclose(ptext);
    
    return gnTotalWordCnt;
}

/*
 * @fn      PRIVATE VOID _UpdateTopKList(UINT32 nNumK, struct wordcount *topk)
 * @param   nNumK
 * @return  topk
 * @par     This function update topk parameter
 * @remark  
 */ 
PRIVATE
VOID _UpdateTopKList(UINT32 nNumK, struct wordcount *topk)
{
    WORD_ENTRY          *pstHashWord = NULL;
    WORD_ENTRY          *pstRankEntity = NULL;
    struct list_head    *pstCurHashList = NULL;
    struct list_head    *pstCurRankList = NULL;
    UINT32              nMinRankRefCnt = 0;
    UINT32              nHashIter, nRankIter;
    
    // Prepare topK list
    pstRankEntity = _GetWordEntity();

    gpstRankListHead = &pstRankEntity->stSlotList;

    for (nRankIter = 0; nRankIter < nNumK - 1; nRankIter++)
    {
        pstRankEntity = _GetWordEntity();
        
        list_add(&pstRankEntity->stSlotList, gpstRankListHead);
    }

    for (nHashIter = 0; nHashIter < NUM_HASH_BUCKET; nHashIter++)
    {
        if (gstWordPool.astTopKBucket[nHashIter].pstSlotHead == NULL)
        {
            continue;
        }
        
        pstCurHashList = gstWordPool.astTopKBucket[nHashIter].pstSlotHead;
        
        while (1)
        {
            pstHashWord = list_entry(pstCurHashList, WORD_ENTRY, stSlotList);

            if (pstHashWord->nTotalRefCnt >= nMinRankRefCnt)
            {
                pstCurRankList = gpstRankListHead;

                for (nRankIter = 0; nRankIter < nNumK; nRankIter++)
                {
                    pstRankEntity = list_entry(pstCurRankList, WORD_ENTRY, stSlotList);

                    if (pstHashWord->nTotalRefCnt >= pstRankEntity->nTotalRefCnt)
                    {
                        pstRankEntity = list_entry(gpstRankListHead->prev, WORD_ENTRY, stSlotList);

                        _CopyString(pstRankEntity->astrWordBuf, pstHashWord->astrWordBuf);
                        pstRankEntity->nTotalRefCnt = pstHashWord->nTotalRefCnt;
                        if (&pstRankEntity->stSlotList != pstCurRankList)
                        {
                            list_move_tail(&pstRankEntity->stSlotList, pstCurRankList);
                        }

                        if (pstCurRankList == gpstRankListHead)
                        {
                            gpstRankListHead = &pstRankEntity->stSlotList;
                        }
                        pstRankEntity = list_entry(gpstRankListHead->prev, WORD_ENTRY, stSlotList);
                        nMinRankRefCnt = pstRankEntity->nTotalRefCnt;
                        break;
                    }

                    pstCurRankList = pstCurRankList->next;
                }
#if 0
                pstCurRankList = gpstRankListHead;
                for (nRankIter = 0; nRankIter < nNumK; nRankIter++)
                {
                    pstRankEntity = list_entry(pstCurRankList, WORD_ENTRY, stSlotList);
                    printf("%3d : %5u %s\n",nRankIter, pstRankEntity->nTotalRefCnt, pstRankEntity->astrWordBuf);
                    pstCurRankList = pstCurRankList->next;
                }
                if (pstCurRankList != gpstRankListHead)
                {
                    while(1);
                }
#endif
            }

            if (pstCurHashList->next == gstWordPool.astTopKBucket[nHashIter].pstSlotHead)
            {
                break;
            }

            // Next word pool
            pstCurHashList = pstCurHashList->next;
        } 
    }

    // Update topk
    pstCurRankList = gpstRankListHead;

    for (nRankIter = 0; nRankIter < nNumK; nRankIter++)
    {
        pstRankEntity = list_entry(pstCurRankList, WORD_ENTRY, stSlotList);
        topk->word = (INT8 *)&pstRankEntity->astrWordBuf;
        topk->count = pstRankEntity->nTotalRefCnt;

        pstCurRankList = pstCurRankList->next;
        topk++;
    }
}
/*
 * @fn      PRIVATE VOID _PushWordPool(WORD_ENTRY *pstWordEntity)
 * @param   pstWordEntity
 * @return  VOID
 * @par     This function implements insert target word in word pool
 * @remark  
 */ 
PRIVATE
VOID _PushWordPool(WORD_ENTRY *pstWordEntity)
{    
    UINT16  nHashBuckeIdx = 0;

    // Get WORD_POOL Key
    nHashBuckeIdx = _GenHashKey(pstWordEntity);

    // Insert to bucket
    _InsertToHashBucket(pstWordEntity, nHashBuckeIdx);
}

/*
 * @fn      PRIVATE UINT16 _GenHashKey(WORD_ENTRY *pstWordEntity)
 * @param   pstWordEntity
 * @return  UINT16 WORD_POOL key value
 * @par     This function generate WORD_POOL key about intput parameter
 * @remark 
 */ 
PRIVATE
UINT16 _GenHashKey(WORD_ENTRY *pstWordEntity)
{
    UINT32 nIter;
    UINT32 *pstWordBuf = (UINT32 *)pstWordEntity->astrWordBuf;
    UINT64 nGenKey = 0;

    for (nIter = 0; nIter < NUM_STRING_CHK_UNIT; nIter++)
    {
        if(*pstWordBuf == 0)
        {
            break;
        }
        else
        {
            nGenKey += *pstWordBuf;
            pstWordBuf++;
        }
    }


    // WORD_POOL key
    return (((nGenKey * nGenKey) >> SHFT_BIT_HASHKEY) & (MASK_BIT_HASHKEY));
}

/*
 * @fn      PRIVATE VOID _InsertToBucket(WORD_ENTRY *pstWordEntity, UINT16 nBucketIdx)
 * @param   pstWordEntity
 * @param   nBucketIdx
 * @return  VOID
 * @par     This function insert word entity to WORD_POOL bucket
 * @remark  If the word entity already exist, it must be destroyed
 */ 
PRIVATE
VOID _InsertToHashBucket(WORD_ENTRY *pstWordEntity, UINT16 nBucketIdx)
{
    // None in slot
    if (gstWordPool.astTopKBucket[nBucketIdx].pstSlotHead == NULL)
    {
        gstWordPool.astTopKBucket[nBucketIdx].pstSlotHead = &pstWordEntity->stSlotList;
        gstWordPool.astTopKBucket[nBucketIdx].nEntryCnt++;
        pstWordEntity->nTotalRefCnt++;
        gnTotalWordPoolCnt++;
    }
    else
    {
        WORD_ENTRY           *pstRefWord = NULL;
        struct list_head    *pstCurList = NULL;
        pstCurList = gstWordPool.astTopKBucket[nBucketIdx].pstSlotHead;
        
        // Search in slot list
        while (1)
        {
            pstRefWord = list_entry(pstCurList, WORD_ENTRY, stSlotList);

            // Check same word
            if (TRUE32 == _CompareString((UINT32 *)pstWordEntity->astrWordBuf, (UINT32 *)pstRefWord->astrWordBuf))
            {
                // Increase reference count
                pstRefWord->nTotalRefCnt++;

                // Adjust slot List in LRU
                if (pstCurList != gstWordPool.astTopKBucket[nBucketIdx].pstSlotHead)
                {
                    list_move_tail(pstCurList, gstWordPool.astTopKBucket[nBucketIdx].pstSlotHead);
                    gstWordPool.astTopKBucket[nBucketIdx].pstSlotHead = pstCurList;
                }
                
                // Not in use, so free this entity
                free(pstWordEntity);

                break;
            }

            // No slot in bucket, add current entity to head
            if (pstCurList->next == gstWordPool.astTopKBucket[nBucketIdx].pstSlotHead)
            {
                // add list head for LRU
                list_add(&pstWordEntity->stSlotList, pstCurList->next);

                gstWordPool.astTopKBucket[nBucketIdx].pstSlotHead = &pstWordEntity->stSlotList;
                gstWordPool.astTopKBucket[nBucketIdx].nEntryCnt++;
                pstWordEntity->nTotalRefCnt++;
                gnTotalWordPoolCnt++;
                break;
            }

            // Next word pool
            pstCurList = pstCurList->next;
        } 
    }
        return;
}

/*
 * @fn      PRIVATE VOID _ReleaseWordPool(VOID)
 * @param   VOID
 * @return  VOID
 * @par     This function release allocated memory of word pool
 * @remark  
 */ 
PRIVATE
VOID _ReleaseWordPool(VOID)
{
    UINT32 nIter;
    WORD_ENTRY *pstWordEntity = NULL;
    struct list_head *pstCurList = NULL;
    struct list_head *pstDelList = NULL;
    UINT32 nEntryCnt = 0;

    for (nIter = 0; nIter < NUM_HASH_BUCKET; nIter++)
    {
        if (gstWordPool.astTopKBucket[nIter].pstSlotHead == NULL)
        {
            continue;
        }

        pstCurList = gstWordPool.astTopKBucket[nIter].pstSlotHead;
        nEntryCnt = gstWordPool.astTopKBucket[nIter].nEntryCnt;
        
        while (nEntryCnt--)
        {
            pstWordEntity = list_entry(pstCurList, WORD_ENTRY, stSlotList);

            // Next word pool
            pstDelList = pstCurList;
            pstCurList = pstCurList->next;
            list_del(pstDelList);
            free(pstWordEntity);
        } 
    }
}

/*
 * @fn      PRIVATE WORD_ENTRY* _GetWordEntity(VOID)
 * @param   VOID
 * @return  WORD_ENTRY* 
 * @par     This function implements allocating word pool
 * @remark 
 */ 
PRIVATE
WORD_ENTRY* _GetWordEntity(VOID)
{
    UINT32 nIter;
    WORD_ENTRY *pstWordEntity = (WORD_ENTRY *)malloc(sizeof(WORD_ENTRY));
    UINT32 *pstrWordBuf = (UINT32 *)pstWordEntity->astrWordBuf;
    INIT_WORD_POOL(pstWordEntity);

    // Clean word buffer
    for (nIter = 0; nIter < NUM_STRING_CHK_UNIT; nIter++, pstrWordBuf++)
    {
        *pstrWordBuf = 0;
    }

    return pstWordEntity;
}

/*
 * @fn      PRIVATE BOOL32 _CheckCriteria(CHAR *pChar)
 * @param   pChar
 * @return  TRUE32      Input parameter is appropriate
 *          FALSE32     Input parameter is illegal
 * @par     This function implements checking appropriate character
 * @remark 
 */ 
PRIVATE
BOOL32 _CheckCriteria(UINT8 *pChar)
{
    BOOL32 bRet = FALSE32;

    // Is it alphabet? (If upper case, it must be changed to lower case.)
    if (((*pChar >= 'A') && (*pChar <= 'Z')) ||
        ((*pChar >= 'a') && (*pChar <= 'z')))
    {
        // change upper case to lower case
        if (*pChar <= 'Z') 
        {
            *pChar = *pChar + 0x20;
        }
        bRet = TRUE32;
    }

    return bRet;
}

/*
 * @fn      PRIVATE BOOL32 _CompareString(UINT32 *pTargetStr, UINT32 *pRefStr)
 * @param   pTargetStr
 * @param   pRefStr
 * @return  TRUE32      string is matched
 *          FALSE32     
 * @par     This function implements comparing string
 * @remark  Warn : this function compare two string in a in 4bytes
 */ 
PRIVATE
BOOL32 _CompareString(UINT32 *pTargetStr, UINT32 *pRefStr)
{
    UINT32 nIter;

    for (nIter = 0; nIter < NUM_STRING_CHK_UNIT; nIter++)
    {
        if ((*pTargetStr == 0) && (*pRefStr == 0))
        {
            break;
        }

        if (*pTargetStr != *pRefStr)
        {
            return FALSE32;
        }

        pTargetStr++;
        pRefStr++;
    }
    
    return TRUE32;
}

/*
 * @fn      PRIVATE VOID _CopyString(UINT8 *pDstStr, UINT8 *pSrcStr)
 * @param   pDstStr
 * @param   pSrcStr
 * @return  VOID
 * @par     This function implements copying string
 * @remark 
 */ 
PRIVATE
VOID _CopyString(UINT8 *pDstStr, UINT8 *pSrcStr)
{
    while(*pSrcStr != '\0')
    {
        *pDstStr = *pSrcStr;

        pDstStr++;
        pSrcStr++;
    }

    *pDstStr = '\0';
}
